@@ -2,11 +2,10 @@ import zmq
 import json
 import os
 from subprocess import Popen
-from platform import system
-from typing import List, Union, Optional, Tuple
+from typing import List, Union, Optional, Tuple, Dict
 from tdw.librarian import ModelLibrarian, SceneLibrarian, MaterialLibrarian, HDRISkyboxLibrarian, \
-    HumanoidAnimationLibrarian, HumanoidLibrarian, HumanoidAnimationRecord
-from tdw.output_data import Version
+    HumanoidAnimationLibrarian, HumanoidLibrarian, HumanoidAnimationRecord, RobotLibrarian
+from tdw.output_data import OutputData, Version
 from tdw.release.build import Build
 from tdw.release.pypi import PyPi
 from tdw.version import __version__
@@ -56,6 +55,7 @@ class Controller(object):
         self.hdri_skybox_librarian: Optional[HDRISkyboxLibrarian] = None
         self.humanoid_librarian: Optional[HumanoidLibrarian] = None
         self.humanoid_animation_librarian: Optional[HumanoidAnimationLibrarian] = None
+        self.robot_librarian: Optional[RobotLibrarian] = None
 
         # Compare the version of the tdw module to the build version.
         if check_version and launch_build:
@@ -70,12 +70,28 @@ class Controller(object):
         :return The output data from the build.
         """
 
-        if not isinstance(commands, list):
-            commands = [commands]
+        if isinstance(commands, list):
+            msg = [json.dumps(commands).encode('utf-8')]
+        else:
+            msg = [json.dumps([commands]).encode('utf-8')]
 
-        self.socket.send_multipart([json.dumps(commands).encode('utf-8')])
+        # Send the commands.
+        self.socket.send_multipart(msg)
+        # Receive output data.
+        resp = self.socket.recv_multipart()
 
-        return self.socket.recv_multipart()
+        # Occasionally, the build's socket will stop receiving messages.
+        # If that happens, it will close the socket, create a new socket, and send a dummy output data object.
+        # The ID of the dummy object is "ftre" (FailedToReceive).
+        # If the controller receives the dummy object, it should re-send its commands.
+        # The dummy object is always in an array: [ftre, 0]
+        # This way, the controller can easily differentiate it from a response that just has the frame count.
+        while len(resp) > 1 and OutputData.get_data_type_id(resp[0]) == "ftre":
+            self.socket.send_multipart(msg)
+            resp = self.socket.recv_multipart()
+
+        # Return the output data from the build.
+        return resp
 
     def start(self, scene="ProcGenScene") -> None:
         """
@@ -204,7 +220,7 @@ class Controller(object):
         :param humanoid_animation_name: The name of the animation.
         :param library: The path to the records file. If left empty, the default library will be selected. See `HumanoidAnimationLibrarian.get_library_filenames()` and `HumanoidAnimationLibrarian.get_default_library()`.
 
-        return An add_humanoid_animation command that the controller can then send.
+        :return An add_humanoid_animation command that the controller can then send.
         """
 
         if self.humanoid_animation_librarian is None:
@@ -215,7 +231,38 @@ class Controller(object):
                 "name": humanoid_animation_name,
                 "url": record.get_url()}, record
 
-    def load_streamed_scene(self, scene="tdw_room_2018") -> None:
+    def get_add_robot(self, name: str, robot_id: int, position: Dict[str, float] = None, rotation: Dict[str, float] = None, library: str = "") -> dict:
+        """
+        Returns a valid add_robot command.
+
+        :param name: The name of the robot.
+        :param robot_id: A unique ID for the robot.
+        :param position: The initial position of the robot. If None, the position will be (0, 0, 0).
+        :param rotation: The initial rotation of the robot in Euler angles.
+        :param library: The path to the records file. If left empty, the default library will be selected. See `RobotLibrarian.get_library_filenames()` and `RobotLibrarian.get_default_library()`.
+
+        :return An `add_robot` command that the controller can then send.
+        """
+
+        if self.robot_librarian is None:
+            self.robot_librarian = RobotLibrarian(library=library)
+
+        record = self.robot_librarian.get_record(name)
+
+        if position is None:
+            position = {"x": 0, "y": 0, "z": 0}
+        if rotation is None:
+            rotation = {"x": 0, "y": 0, "z": 0}
+
+        assert record is not None, f"Robot metadata record not found: {name}"
+        return {"$type": "add_robot",
+                "id": robot_id,
+                "position": position,
+                "rotation": rotation,
+                "name": name,
+                "url": record.get_url()}
+
+    def load_streamed_scene(self, scene="tdw_room") -> None:
         """
         Load a streamed scene. This is equivalent to: `c.communicate(c.get_add_scene(scene))`
 
