@@ -36,6 +36,11 @@ class TDWUtils:
 
     VECTOR3_ZERO = {"x": 0, "y": 0, "z": 0}
 
+    # Cached values used during point cloud generation.
+    __WIDTH: int = -1
+    __HEIGHT: int = -1
+    __CAM_TO_IMG_MAT: Optional[np.array] = None
+
     @staticmethod
     def vector3_to_array(vector3: Dict[str, float]) -> np.array:
         """
@@ -278,8 +283,30 @@ class TDWUtils:
                 TDWUtils.get_pil_image(images, i).resize((resize_to[0], resize_to[1]), Image.LANCZOS)\
                     .save(os.path.join(output_directory, fi))
             else:
-                with open(os.path.join(output_directory, fi), "wb") as f:
-                    f.write(images.get_image(i))
+                pass_mask = images.get_pass_mask(i)
+                path = os.path.join(output_directory, fi)
+                # The depth passes aren't png files, so we need to convert them.
+                if pass_mask == "_depth" or pass_mask == "_depth_simple":
+                    # Save the image.
+                    Image.fromarray(TDWUtils.get_shaped_depth_pass(images=images, index=i)).save(path)
+                # Every other pass can be saved directly to disk.
+                else:
+                    with open(path, "wb") as f:
+                        f.write(images.get_image(i))
+
+    @staticmethod
+    def get_shaped_depth_pass(images: Images, index: int) -> np.array:
+        """
+        The `_depth` and `_depth_simple` passes are a 1D array of RGB values, as oppposed to a png or jpg like every other pass.
+        This function reshapes the array into a 2D array of RGB values.
+
+        :param images: The `Images` output data.
+        :param index: The index in `Images` of the depth pass. See: `Images.get_pass_mask()`.
+
+        :return: A reshaped depth pass. Shape is: `(height, width, 3)`.
+        """
+
+        return np.flip(np.reshape(images.get_image(index), (images.get_height(), images.get_width(), 3)), 0)
 
     @staticmethod
     def zero_padding(integer: int, width=4) -> str:
@@ -367,23 +394,99 @@ class TDWUtils:
         return commands
 
     @staticmethod
-    def get_depth_values(image: np.array) -> np.array:
+    def get_depth_values(image: np.array, depth_pass: str = "_depth", width: int = 256, height: int = 256, near_plane: float = 0.1, far_plane: float = 100) -> np.array:
         """
         Get the depth values of each pixel in a _depth image pass.
         The far plane is hardcoded as 100. The near plane is hardcoded as 0.1.
         (This is due to how the depth shader is implemented.)
 
         :param image: The image pass as a numpy array.
+        :param depth_pass: The type of depth pass. This determines how the values are decoded. Options: `"_depth"`, `"_depth_simple"`.
+        :param width: The width of the screen in pixels. See output data `Images.get_width()`.
+        :param height: The height of the screen in pixels. See output data `Images.get_height()`.
+        :param near_plane: The near clipping plane. See command `set_camera_clipping_planes`. The default value in this function is the default value of the near clipping plane.
+        :param far_plane: The far clipping plane. See command `set_camera_clipping_planes`. The default value in this function is the default value of the far clipping plane.
 
         :return An array of depth values.
         """
 
         # Convert the image to a 2D image array.
-        image = np.array(Image.open(io.BytesIO(image)), dtype=float)
+        image = np.flip(np.reshape(image, (height, width, 3)), 0)
+        if depth_pass == "_depth":
+            depth_values = np.array((image[:, :, 0] + image[:, :, 1] / 256.0 + image[:, :, 2] / (256.0 ** 2)))
+        elif depth_pass == "_depth_simple":
+            depth_values = image[:, :, 0] / 256.0
+        else:
+            raise Exception(f"Invalid depth pass: {depth_pass}")
+        # Un-normalize the depth values.
+        return (depth_values * ((far_plane - near_plane) / 256.0)).astype(np.float32)
 
-        depth = np.array((image[:, :, 0] * 256 * 256 + image[:, :, 1] * 256 + image[:, :, 2]) /
-                         (256 * 256 * 256) * 100.1)
-        return depth
+    @staticmethod
+    def get_point_cloud(depth, camera_matrix: Union[np.array, tuple], vfov: float = 54.43222, filename: str = None, near_plane: float = 0.1, far_plane: float = 100) -> np.array:
+        """
+        Create a point cloud from an numpy array of depth values.
+
+        :param depth: Depth values converted from a depth pass. See: `TDWUtils.get_depth_values()`
+        :param camera_matrix: The camera matrix as a tuple or numpy array. See: [`send_camera_matrices`](https://github.com/threedworld-mit/tdw/blob/master/Documentation/api/command_api.md#send_camera_matrices).
+        :param vfov: The field of view. See: [`set_field_of_view`](https://github.com/threedworld-mit/tdw/blob/master/Documentation/api/command_api.md#set_field_of_view)
+        :param filename: If not None, the point cloud data will be written to this file.
+        :param near_plane: The near clipping plane. See command `set_camera_clipping_planes`. The default value in this function is the default value of the near clipping plane.
+        :param far_plane: The far clipping plane. See command `set_camera_clipping_planes`. The default value in this function is the default value of the far clipping plane.
+
+        :return: An point cloud as a numpy array of `[x, y, z]` coordinates.
+        """
+
+        if isinstance(camera_matrix, tuple):
+            camera_matrix = np.array(camera_matrix)
+        camera_matrix = np.linalg.inv(camera_matrix.reshape((4, 4)))
+
+        # Different from real-world camera coordinate system.
+        # OpenGL uses negative z axis as the camera front direction.
+        # x axes are same, hence y axis is reversed as well.
+        # Source: https://learnopengl.com/Getting-started/Camera
+        rot = np.array([[1, 0, 0, 0],
+                        [0, -1, 0, 0],
+                        [0, 0, -1, 0],
+                        [0, 0, 0, 1]])
+        camera_matrix = np.dot(camera_matrix, rot)
+
+        # Cache some calculations we'll need to use every time.
+        if TDWUtils.__HEIGHT != depth.shape[0] or TDWUtils.__WIDTH != depth.shape[1]:
+            TDWUtils.__HEIGHT = depth.shape[0]
+            TDWUtils.__WIDTH = depth.shape[1]
+
+            img_pixs = np.mgrid[0: depth.shape[0], 0: depth.shape[1]].reshape(2, -1)
+            # Swap (v, u) into (u, v).
+            img_pixs[[0, 1], :] = img_pixs[[1, 0], :]
+            img_pix_ones = np.concatenate((img_pixs, np.ones((1, img_pixs.shape[1]))))
+
+            # Calculate the intrinsic matrix from vertical_fov.
+            # Motice that hfov and vfov are different if height != width
+            # We can also get the intrinsic matrix from opengl's perspective matrix.
+            # http://kgeorge.github.io/2014/03/08/calculating-opengl-perspective-matrix-from-opencv-intrinsic-matrix
+            vfov = vfov / 180.0 * np.pi
+            tan_half_vfov = np.tan(vfov / 2.0)
+            tan_half_hfov = tan_half_vfov * TDWUtils.__WIDTH / float(TDWUtils.__HEIGHT)
+            fx = TDWUtils.__WIDTH / 2.0 / tan_half_hfov  # focal length in pixel space
+            fy = TDWUtils.__HEIGHT / 2.0 / tan_half_vfov
+            intrinsics = np.array([[fx, 0, TDWUtils.__WIDTH / 2.0],
+                                   [0, fy, TDWUtils.__HEIGHT / 2.0],
+                                   [0, 0, 1]])
+            img_inv = np.linalg.inv(intrinsics[:3, :3])
+            TDWUtils.__CAM_TO_IMG_MAT = np.dot(img_inv, img_pix_ones)
+
+        points_in_cam = np.multiply(TDWUtils.__CAM_TO_IMG_MAT, depth.reshape(-1))
+        points_in_cam = np.concatenate((points_in_cam, np.ones((1, points_in_cam.shape[1]))), axis=0)
+        points_in_world = np.dot(camera_matrix, points_in_cam)
+        points_in_world = points_in_world[:3, :].reshape(3, TDWUtils.__WIDTH, TDWUtils.__HEIGHT)
+        points_in_cam = points_in_cam[:3, :].reshape(3, TDWUtils.__WIDTH, TDWUtils.__HEIGHT)
+        if filename is not None:
+            f = open(filename, 'w')
+            for i in range(points_in_world.shape[1]):
+                for j in range(points_in_world.shape[2]):
+                    if points_in_cam[2, i, j] < (far_plane - near_plane):
+                        f.write(f'{points_in_world[0, i, j]};{points_in_world[1, i, j]};{points_in_world[2, i, j]}\n')
+        return points_in_world
 
     @staticmethod
     def create_avatar(avatar_type="A_Img_Caps_Kinematic", avatar_id="a", position=None, look_at=None) -> List[dict]:
@@ -610,24 +713,6 @@ class TDWUtils:
 
         return np.array([(hashable >> 16) & 255, (hashable >> 8) & 255, hashable & 255], dtype=int)
 
-    @staticmethod
-    def get_collisions(resp: List[bytes]) -> Tuple[List[Collision], List[EnvironmentCollision]]:
-        """
-        :param resp: The response from the build (a byte array).
-
-        :return: Tuple: A list of collisions; a list of environment collisions.
-        """
-
-        collisions: List[Collision] = list()
-        env_collisions: List[EnvironmentCollision] = list()
-        for i in range(len(resp) - 1):
-            r_id = OutputData.get_data_type_id(resp[i])
-            if r_id == "coll":
-                collisions.append(Collision(resp[i]))
-            elif r_id == "enco":
-                env_collisions.append(EnvironmentCollision(resp[i]))
-        return collisions, env_collisions
-
     @staticmethod
     def get_bounds_dict(bounds: Bounds, index: int) -> Dict[str, np.array]:
         """
@@ -835,7 +920,38 @@ class QuaternionUtils:
     Vectors are always numpy arrays in the following order: `[x, y, z]`.
     """
 
-    _UP = np.array([0, 1, 0])
+    """:class_var
+    The global up directional vector.
+    """
+    UP = np.array([0, 1, 0])
+    """:class_var
+    The global forward directional vector.
+    """
+    FORWARD: np.array = np.array([0, 0, 1])
+    """:class_var
+    The quaternion identity rotation.
+    """
+    IDENTITY = np.array([0, 0, 0, 1])
+
+    @staticmethod
+    def get_inverse(q: np.array) -> np.array:
+        """
+        Source: https://referencesource.microsoft.com/#System.Numerics/System/Numerics/Quaternion.cs
+
+        :param q: The quaternion.
+
+        :return: The inverse of the quaternion.
+        """
+
+        x = q[0]
+        y = q[1]
+        z = q[2]
+        w = q[3]
+
+        ls = x * x + y * y + z * z + w * w
+        inv = 1.0 / ls
+
+        return np.array([-x * inv, -y * inv, -z * inv, w * inv])
 
     @staticmethod
     def multiply(q1: np.array, q2: np.array) -> np.array:
@@ -884,6 +1000,7 @@ class QuaternionUtils:
     @staticmethod
     def multiply_by_vector(q: np.array, v: np.array) -> np.array:
         """
+        Multiply a quaternion by a vector.
         Source: https://stackoverflow.com/questions/4870393/rotating-coordinate-system-via-a-quaternion
 
         :param q: The quaternion.
@@ -895,6 +1012,21 @@ class QuaternionUtils:
         q2 = (v[0], v[1], v[2], 0.0)
         return QuaternionUtils.multiply(QuaternionUtils.multiply(q, q2), QuaternionUtils.get_conjugate(q))[:-1]
 
+    @staticmethod
+    def world_to_local_vector(position: np.array, origin: np.array, rotation: np.array) -> np.array:
+        """
+        Convert a vector position in absolute world coordinates to relative local coordinates.
+        Source: https://answers.unity.com/questions/601062/what-inversetransformpoint-does-need-explanation-p.html
+
+        :param position: The position vector in world coordinates.
+        :param origin: The origin vector of the local space in world coordinates.
+        :param rotation: The rotation quaternion of the local coordinate space.
+
+        :return: `position` in local coordinates.
+        """
+
+        return QuaternionUtils.multiply_by_vector(q=QuaternionUtils.get_inverse(q=rotation), v=position - origin)
+
     @staticmethod
     def get_up_direction(q: np.array) -> np.array:
         """
@@ -903,7 +1035,7 @@ class QuaternionUtils:
         :return: A directional vector corresponding to the "up" direction from the quaternion.
         """
 
-        return QuaternionUtils.multiply_by_vector(q, QuaternionUtils._UP)
+        return QuaternionUtils.multiply_by_vector(q, QuaternionUtils.UP)
 
     @staticmethod
     def euler_angles_to_quaternion(euler: np.array) -> np.array:
